#!/usr/bin/python
#-*- coding: utf-8 -*-
# ---------------------------------------------------------------------------
# This file is a plugin for anki flashcard application http://repose.cx/anki/
# ---------------------------------------------------------------------------
# File:        tonemark.py
# Description: Converts numbered pinyin to tonemark pinyin for comparison
#
# Author:      Eric Grimm (effectivechinese@gmail.com)
# Version:     0.01 (2009-04-23)
# License:     GPL
# ---------------------------------------------------------------------------
# Changelog:
# ---- 0.01 -- 2009-04-23 -- Eric Grimm  ----
#   initial release
# ---------------------------------------------------------------------------

from PyQt4.QtCore import *
from anki.sound import playFromText
from anki.utils import stripHTML
from ankiqt.ui.utils import mungeQA
from ankiqt.ui.view import failedCharColour, passedCharColour, View

import re

from ankiqt import mw

EOL = u'\n'
TAB = u'\t'

cedict = {}
punctuation = [u' ', u'.', u',', u'?', u';', u'!']

cedict_entry = re.compile(
    r'''(?ix)                           # ignore case, verbose definitions
    (\s*)                               # simplified character
    (\s*)                               # traditional character
    [                                   # left bracket
    (\s*)                               # pinyin with tone numbers
    ]                                   # right bracket
    (\s*)                               # english definition   
    ''')

_UMLAUT_U = u'\xfc'

# tone number to combining diacritical table. 
_TONE = {u'1':u'\u0304',
        u'2':u'\u0301',
        u'3':u'\u030c',
        u'4':u'\u0300',
        u'5':u''}

# Note: The following algorithm was contributed by Mark Tolonen
#
def _convert(matchobj):
    """_convert(matchobj) -> replacement text
    Helper function for re.sub with three groups.
    Group 1 - one or more vowels
    Group 2 - zero or more consonants
    Group 3 - tone character '1'-'5'
    Applies pinyin tone marks to the appropriate vowel.
    Algorithm:
    In vowels, replace 'v' with umlaut-u.
    First 'a' or 'e' in vowels gets the mark,
    else first 'ou', 'o' gets the mark,
    else last vowel gets the mark.
    """
    import unicodedata
    vowels,consonants,tone = matchobj.groups()
    vowels = vowels.replace(u'v',_UMLAUT_U)
    n = vowels.lower().find(u'a')
    if n == -1:
        n = vowels.lower().find(u'e')
    if n == -1:
        n = vowels.lower().find(u'ou')
    if n != -1:
        # a, e, or ou found.
        # Add accent after the a, e, or o
        retval = vowels[:n+1] + _TONE[tone] + vowels[n+1:] + consonants
    else:
        # accent after the last vowel
        retval = vowels + _TONE[tone] + consonants
    return unicodedata.normalize('NFC',retval)

def Convert(pinyin_with_tone_numbers):
    """Convert(pinyin_with_tone_numbers) -> pinyin_with_tone_marks
    Given a pinyin string with tone numbers, return a converted unicode
    string with tone marks.  Requires 'v' to represent umlaut-u.
    """
    import re
    # Calls conversion function for each match.
    return re.sub(
        u'''(?ix)       # ignore case, verbose
        ([aeiouv]+)     # one or more vowels (v for umlaut-u)
        ([^aeiouv]*)    # zero or more non-vowels
        ([12345])       # tone mark 1-5
        ''',_convert,pinyin_with_tone_numbers)

    
def drawAnswerDC():
    "Show the answer."
    view = mw.bodyView

    #Convert the answer to tone marked pinyin
    input = unicode(view.main.typeAnswerField.text())
    view.main.typeAnswerField.setText(Convert(input))

    #Call the real function
    View.drawAnswer(view)

def initConvertToneMarks():
    "Overwrite drawAnswer"
    mw.bodyView.drawAnswer = drawAnswerDC

mw.addHook("init", initConvertToneMarks)
