# coding: utf-8

# Copyright 2011 Robert Siemer <Robert.Siemer-anki@backsla.sh>

# This file is part of “Panki Chinese”.

# “Panki Chinese” is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# “Panki Chinese” is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with “Panki Chinese”.  If not, see <http://www.gnu.org/licenses/>.



# Main plugin file for “Panki Chinese”. It patches Anki 1.2.5 and 1.2.7 from
# “inside” (only while enabled) in
# order to enable it to run the corresponding deck. The deck is not included.
# See  http://backsla.sh/anki/panki  for instructions on how to get it.

# The main file does:
# -disables the use and generation of “anki deck caches”
# -changes card template mechanisms
# -changes audio playback behavior
#  +the audio changes are skipped on Windows XP (it hangs when it tries playing)

# To form a complete “Panki Chinese”, the deck and two other plugins are needed:
# CardHistory.py and PankiTemplates.py

# While this plugin allows the aforementioned plugins to work, it leaves some
# parts of Anki broken without fixing (I’m to lazy to fix): latex, parts of
# the card browser, media handling (except local media).

# Before using other decks, switch the PankiChinese plugin off and restart Anki.
# The other plugins (CardHistory, PankiTemplate) can
# stay on, they are programmed to fall back to do nothing by themselves.


import os, subprocess, re, platform
import unicodedata as ucd
import ankiqt, anki
from anki.utils import stripHTML
from anki.lang import _
from anki.sound import playFromText, clearAudioQueue, stripSounds
from anki.sound import mplayerCmd, mplayerEvt, si, MplayerMonitor
from anki.hooks import addHook, removeHook, _hooks
from anki.media import stripMedia
from anki.template import render
from PyQt4.QtCore import QUrl, SIGNAL


# how to convert a patch to a plugin:
# 1) all functions assigned/re-registered?
# 2) all external variables here imported?
# 3) no collision between external variables and function names here?
# 4) “foreign” globals rewired to originating variables? (Otherwise I create a copy...)
# 5) assure no reassigned function got imported before via “from ... import FUNC”
#    or overwrite these, too


returnNone = lambda *p, **kw: None



ankiqt.mw.mainWin.actionRepeatAudio.setShortcut(_("F3"))



ankiqt.ui.cardlist.EditDeck.rebuildSortIndex = returnNone



def formatChanged(self, type):
    if self.updatingCards:
        return
    if type == "question":
        text = unicode(self.form.cardQuestion.toPlainText())
        text = self.screenToFormat(text)
        #self.realCardModel(self.card).qformat = text
        self.card.cardModel.qformat = text
    else:
        text = unicode(self.form.cardAnswer.toPlainText())
        text = self.screenToFormat(text)
        self.card.cardModel.aformat = text
    self.fact.model.setModified()
    self.deck.flushMod()
    self.deck.setModified()
    self.needFormatRebuild = True
    self.renderPreview()

def renderPreview(self):
    c = self.card
    self.form.preview.setHtml(
        runFilter("drawQuestion", stripSounds(c.getQuestion()), c) + "<hr>" +
        runFilter("drawAnswer", stripSounds(c.getAnswer()), c),
        self.deck.mediaQUrl())
    clearAudioQueue()
    if c.id not in self.playedAudio:
        playFromText(c.getQuestion() + c.getAnswer())
        self.playedAudio[c.id] = True

ankiqt.ui.clayout.CardLayout.formatChanged = formatChanged
ankiqt.ui.clayout.CardLayout.renderPreview = renderPreview



def onCardAnsweredHook(self, cardId, isLeech):
    if not isLeech:
        self.setNotice()
        return
    txt = (_("""\
<b>%s</b>... is a <a href="http://ichi2.net/anki/wiki/Leeches">leech</a>.""")
            % stripHTML(stripSounds(self.currentCard.getQuestion())).\
            replace("\n", " ")[0:30])
    if isLeech and self.deck.s.scalar(
        "select 1 from cards where id = :id and priority < 1", id=cardId):
        txt += _(" It has been suspended.")
    self.setNotice(txt)

def enableCardMenuItems(self):
    self.maybeEnableUndo()
    self.mainWin.actionEditLayout.setEnabled(True)
    self.mainWin.actionRepeatAudio.setEnabled(True)
    self.mainWin.actionMarkCard.setEnabled(True)
    self.mainWin.actionSuspendCard.setEnabled(True)
    self.mainWin.actionDelete.setEnabled(True)
    self.mainWin.actionBuryFact.setEnabled(True)
    enableEdits = (not self.config['preventEditUntilAnswer'] or
                    self.state != "getQuestion")
    self.mainWin.actionEditCurrent.setEnabled(enableEdits)
    self.mainWin.actionEditdeck.setEnabled(enableEdits)
    runHook("enableCardMenuItems")

def onRepeatAudio(self):
    clearAudioQueue()
    audio = u''
    if (not self.currentCard.cardModel.questionInAnswer
        or self.state == "showQuestion") and \
        self.config['repeatQuestionAudio']:
        audio += self.currentCard.getQuestion()
    if self.state != "showQuestion":
        audio += self.currentCard.getAnswer()
    playFromText(audio)

removeHook('cardAnswered', ankiqt.mw.onCardAnsweredHook)
ankiqt.ui.main.AnkiQt.onCardAnsweredHook = onCardAnsweredHook
addHook("cardAnswered", ankiqt.mw.onCardAnsweredHook)
ankiqt.ui.main.AnkiQt.enableCardMenuItems = enableCardMenuItems
assert ankiqt.mw.disconnect(ankiqt.mw.mainWin.actionRepeatAudio, SIGNAL('triggered()'),
    ankiqt.mw.onRepeatAudio)
ankiqt.ui.main.AnkiQt.onRepeatAudio = onRepeatAudio
assert ankiqt.mw.connect(ankiqt.mw.mainWin.actionRepeatAudio, SIGNAL('triggered()'),
    ankiqt.mw.onRepeatAudio)



def drawQuestion(self, nosound=False):
    "Show the question."
    if not self.main.config['splitQA']:
        self.write("<br>")
    q = self.main.currentCard.getQuestion()
    q = runFilter("drawQuestion", q, self.main.currentCard)
    self.write(stripSounds(q))
    if (self.state != self.oldState and not nosound
        and self.main.config['autoplaySounds']):
        clearAudioQueue()
        playFromText(q)
    if self.main.currentCard.cardModel.typeAnswer and hasattr(self, 'adjustInputFont'):
        self.adjustInputFont()

def flush(self):
    "Write the current HTML buffer to the screen."
    runHook("preFlushHook")
    self.body.setHtml(self.buffer, self.main.deck.mediaQUrl())

def drawAnswer(self):
    "Show the answer."
    a = self.main.currentCard.getAnswer()
    a = runFilter("drawAnswer", a, self.main.currentCard)
    if self.main.currentCard.cardModel.typeAnswer:
        try:
            cor = stripMedia(stripHTML(self.main.currentCard.fact[
                self.main.currentCard.cardModel.typeAnswer]))
        except KeyError:
            self.main.currentCard.cardModel.typeAnswer = ""
            cor = ""
        if cor:
            given = unicode(self.main.typeAnswerField.text())
            res = self.correct(ucd.normalize('NFC', cor),
                                ucd.normalize('NFC', given))
            a = res + "<br>" + a
    self.write(stripSounds(a))
    if self.state != self.oldState and self.main.config['autoplaySounds']:
        playFromText(a)

# just a self.flush() less (not needed, and my flush() needs a deck)
def setState(self, state):
    "Change to STATE, and update the display."
    self.oldState = getattr(self, 'state', None)
    self.state = state
    if self.state == "initial":
        return
    elif self.state == "noDeck":
        self.clearWindow()
        self.drawWelcomeMessage()
        return
    self.redisplay()

ankiqt.ui.view.View.drawQuestion = drawQuestion
ankiqt.ui.view.View.flush = flush
ankiqt.ui.view.View.drawAnswer = drawAnswer
ankiqt.ui.view.View.setBackground = returnNone
ankiqt.ui.view.View.setState = setState
del ankiqt.ui.view.View.mungeQA
del ankiqt.ui.view.View.addStyles
del ankiqt.ui.view.View.center



def questionanswer(self, which):
    fields = {}
    for f in self.fact.model.fieldModels:
        value = self.fact[f.name]
        #fields["text:" + f.name] = stripHTML(value)
        if value:
            fields[f.name] = '<span class="%s">%s</span>' % (cleanName(f.name), value)
        else:
            fields[f.name] = u""
    fields['tags'] = self.fact.tags
    fields['Tags'] = self.fact.tags
    fields['modelTags'] = self.fact.model.tags
    fields['cardModel'] = self.cardModel.name
    format = self.cardModel.qformat if which == 'question' else self.cardModel.aformat
    # convert old style
    #format = re.sub("%\((.+?)\)s", "{{\\1}}", format)
    # allow custom rendering functions & info
    kwargs = dict(qa=which, card=self)
    fields = runFilter("prepareFields", fields, **kwargs)
    format = runFilter("preRenderFormatQA", format, **kwargs)
    html = render(format, fields)
    html = runFilter("formatQA", html, **kwargs)
    return html

def getQuestion(self):
    return self.questionanswer('question')

def getAnswer(self):
    return self.questionanswer('answer')

def fromDB(self, s, id):
    r = s.first("""SELECT
        id, factId, cardModelId, created, modified, tags, ordinal,
        priority, interval, lastInterval, due, lastDue, factor,
        lastFactor, firstAnswered, reps, successive, averageTime, reviewTime,
        youngEase0, youngEase1, youngEase2, youngEase3, youngEase4,
        matureEase0, matureEase1, matureEase2, matureEase3, matureEase4,
        yesCount, noCount, spaceUntil, isDue, type, combinedDue
        FROM cards WHERE id = :id""", id=id)
    if not r:
        return
    (self.id,
        self.factId,
        self.cardModelId,
        self.created,
        self.modified,
        self.tags,
        self.ordinal,
        self.priority,
        self.interval,
        self.lastInterval,
        self.due,
        self.lastDue,
        self.factor,
        self.lastFactor,
        self.firstAnswered,
        self.reps,
        self.successive,
        self.averageTime,
        self.reviewTime,
        self.youngEase0,
        self.youngEase1,
        self.youngEase2,
        self.youngEase3,
        self.youngEase4,
        self.matureEase0,
        self.matureEase1,
        self.matureEase2,
        self.matureEase3,
        self.matureEase4,
        self.yesCount,
        self.noCount,
        self.spaceUntil,
        self.isDue,
        self.type,
        self.combinedDue) = r
    return True

anki.cards.Card.questionanswer = questionanswer
anki.cards.Card.getQuestion = getQuestion
anki.cards.Card.getAnswer = getAnswer
anki.cards.Card.htmlAnswer = anki.cards.Card.getAnswer
anki.cards.Card.fromDB = fromDB
anki.cards.Card.rebuildQA = returnNone
del anki.cards.Card.htmlQuestion
del anki.cards.Card.splitTags


def mediaQUrl(self):
    return QUrl.fromLocalFile(self.mediaDir(create=None) + os.sep)

def progressHandler(self):
    if self.progressHandlerEnabled:
        runHook("dbProgress")

# only if lower the call rate as well (I skipped that here)!
# anki.deck.Deck.progressHandler = progressHandler
anki.deck.Deck.mediaQUrl = mediaQUrl
anki.deck.Deck.rebuildCSS = returnNone
anki.deck.Deck.updateCardQACacheFromIds = returnNone
anki.deck.Deck.updateCardsFromModel = returnNone
anki.deck.Deck.updateDynamicIndices = returnNone
anki.deck.Deck.updateFieldCache = returnNone
anki.deck.DeckStorage._addIndices = staticmethod(returnNone)
del anki.deck.Deck.updateCardQACache


def runHook(hook, *args, **kwargs):
    "Run all functions on hook."
    hook = _hooks.get(hook, None)
    if hook:
        for func in hook:
            func(*args, **kwargs)

def runFilter(hook, arg, *args, **kwargs):
    hook = _hooks.get(hook, None)
    if hook:
        for func in hook:
            arg = func(arg, *args, **kwargs)
    return arg

anki.hooks.runHook = runHook
anki.hooks.runFilter = runFilter

ankiqt.ui.view.runFilter = runFilter
ankiqt.ui.facteditor.runFilter = runFilter
ankiqt.ui.clayout.runFilter = runFilter
anki.stats.runFilter = runFilter
anki.models.runFilter = runFilter
ankiqt.runHook = runHook
ankiqt.ui.help.runHook = runHook
ankiqt.ui.view.runHook = runHook
ankiqt.ui.facteditor.runHook = runHook
ankiqt.ui.cardlist.runHook = runHook
ankiqt.ui.main.runHook = runHook
anki.db.runHook = runHook
anki.sync.runHook = runHook
anki.facts.runHook = runHook
anki.sound.runHook = runHook
anki.deck.runHook = runHook



removeHook('formatQA', anki.latex.formatQA)



del anki.models.formatQA



def run(self):
    self.mplayer = None
    self.deadPlayers = []
    while 1:
        mplayerEvt.wait()
        if anki.sound.mplayerClear:
            self.commandMplayer('stop')
            anki.sound.mplayerQueue = []
            anki.sound.mplayerClear = False
        while anki.sound.mplayerQueue:
            item = anki.sound.mplayerQueue.pop(0)
            self.commandMplayer('loadfile "%s" 1' % (item))
        # wait() on finished processes. we don't want to block on the
        # wait, so we keep trying each time we're reactivated
        def clean(pl):
            if pl.poll() is not None:
                pl.wait()
                return False
            else:
                return True
        self.deadPlayers = [pl for pl in self.deadPlayers if clean(pl)]
        mplayerEvt.clear()

def startProcess(self):
    try:
        cmd = mplayerCmd + ["-slave", "-idle"]
        self.mplayer = subprocess.Popen(
            cmd, startupinfo=si, stdin=subprocess.PIPE,
            stderr=file(os.devnull, 'w'))
    except OSError:
        mplayerEvt.clear()
        raise Exception("Audio player not found")

def commandMplayer(self, cmd):
    cmd += '\n'
    if not self.mplayer:
        self.startProcess()
    try:
        self.mplayer.stdin.write(cmd)
    except:
        # mplayer has quit and needs restarting
        self.deadPlayers.append(self.mplayer)
        self.mplayer = None
        self.startProcess()
        self.mplayer.stdin.write(cmd)

def ensureMplayerThreads():
    if not anki.sound.mplayerManager:
        anki.sound.mplayerManager = MplayerMonitor()
        anki.sound.mplayerManager.daemon = True
        anki.sound.mplayerManager.start()

if platform.win32_ver()[0] != 'XP':
    anki.sound.MplayerMonitor.run = run
    anki.sound.MplayerMonitor.startProcess = startProcess
    anki.sound.MplayerMonitor.commandMplayer = commandMplayer
    anki.sound.ensureMplayerThreads = ensureMplayerThreads
    del anki.sound.MplayerReader



_cleanNameRe = re.compile(r'[\\ "&]')
def cleanName(name):
    'remove space, backslash, double quote and ampersand'
    return _cleanNameRe.sub('', name)

# anki.utils.cleanName = cleanName

from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import mapper, clear_mappers, properties
import sqlalchemy.orm as orm
relationship = orm.relationship if hasattr(orm, 'relationship') else orm.relation

clear_mappers()
mapper(anki.history.CardHistoryEntry, anki.history.reviewHistoryTable)
mapper(anki.models.FieldModel, anki.models.fieldModelsTable)
mapper(anki.models.CardModel, anki.models.cardModelsTable)
mapper(anki.models.Model, anki.models.modelsTable, properties={
    'fieldModels': relationship(anki.models.FieldModel, backref='model',
                             collection_class=ordering_list('ordinal'),
                             order_by=[anki.models.fieldModelsTable.c.ordinal],
                            cascade="all, delete-orphan"),
    'cardModels': relationship(anki.models.CardModel, backref='model',
                           collection_class=ordering_list('ordinal'),
                           order_by=[anki.models.cardModelsTable.c.ordinal],
                           cascade="all, delete-orphan"),
       })
mapper(anki.stats.Stats, anki.stats.statsTable)
mapper(anki.media.Media, anki.media.mediaTable)
mapper(anki.facts.Field, anki.facts.fieldsTable, properties={
    'fieldModel': relationship(anki.models.FieldModel)
    })
mapper(anki.cards.Card, anki.cards.cardsTable, properties={
    'cardModel': relationship(anki.models.CardModel),
    'fact': relationship(anki.facts.Fact, backref="cards", primaryjoin=
                     anki.cards.cardsTable.c.factId == anki.facts.factsTable.c.id),
    })
mapper(anki.facts.Fact, anki.facts.factsTable, properties={
    'model': relationship(anki.models.Model),
    'fields': relationship(anki.facts.Field, backref="fact",
                            order_by=anki.facts.Field.ordinal),
    })
mapper(anki.deck.Deck, anki.deck.decksTable, properties={
    'currentModel': relationship(anki.models.Model, primaryjoin=
                                anki.deck.decksTable.c.currentModelId ==
                                anki.models.modelsTable.c.id),
    'models': relationship(anki.models.Model, post_update=True, backref='deck',
                        primaryjoin=
                            anki.deck.decksTable.c.id ==
                            anki.models.modelsTable.c.deckId),
    })
